# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

from copy import deepcopy

import torch
import torch.nn as nn

from ._quantize import fp8_quantize
from ._quantize_pertensor import fp8_quantize_pertensor


class Coat_quantize_bgn(nn.Module):
    def __init__(self, args=None, layer_type=""):
        super().__init__()
        self.args = deepcopy(args)
        self.fp8type = self.args.fabit
        self.layer_type = layer_type

    def forward(self, input):
        if self.training:
            return Coat_quantize_bgn_func.apply(input, self.args.group_size, self.fp8type)
        else:
            return input, None, None


class Coat_quantize_bgn_func(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, group_size, fp8type):
        """
        (Qoutput, Oscale) uses 1 * 16 quantization
        """
        Qoutput, Oscale = fp8_quantize(input, group_size, fp8type)
        # For autograd
        Qoutput = Qoutput.view(torch.float8_e4m3fn)
        ctx.saved = group_size
        return input, Qoutput, Oscale

    @staticmethod
    def backward(ctx, grad_output, Qgrad_output, Gscale):
        """
        (Qgrad_output, Gscale) uses 1 * 16 quantization
        """
        return grad_output, None, None


class Coat_quantize_end(nn.Module):
    def __init__(self, args=None, layer_type=""):
        super().__init__()
        self.args = deepcopy(args)
        self.fp8type = self.args.babit
        self.layer_type = layer_type

    def forward(self, input, Qinput, Iscale):
        if self.training:
            return Coat_quantize_end_func.apply(input, Qinput, Iscale, self.args.group_size, self.fp8type)
        else:
            return input


class Coat_quantize_end_func(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, Qinput, Iscale, group_size, fp8type):
        """
        (Qinput, Iscale) uses 1 * 16 quantization
        """
        ctx.saved = group_size, fp8type

        return input

    @staticmethod
    def backward(ctx, grad_output):
        """
        (Qgrad_output, Gscale) uses per-tensor quantization
        """

        group_size, fp8type = ctx.saved
        Qgrad_output, Gscale, Gscale_g16 = fp8_quantize_pertensor(grad_output, group_size, fp8type, stochastic=False)

        # For autograd
        Qgrad_output = Qgrad_output.view(torch.float8_e4m3fn)

        return grad_output, Qgrad_output, Gscale_g16, None, None



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

"""
Layer Normalization
====================
In this tutorial, you will write a high-performance layer normalization
kernel that runs faster than the PyTorch implementation.

In doing so, you will learn about:

* Implementing backward pass in Triton.

* Implementing parallel reduction in Triton.

"""

import torch
import triton
import triton.language as tl

from ._division_transpose import fp8_division_transpose
from .common import FP8_MAX_VALUE, SCALE_MIN_THRES

"""RMSNorm Forward + Backward"""
"""Forward: Input uses 1 * 16 group quantization"""
"""Forward: Output uses per-tensor quantization"""
"""Backward: Input uses full-precision/BF16."""
"""Backward: Output uses full-precision/BF16."""

"""The input can be 2D or 3D, but the calculation is performed in 2D"""


@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["N"] // args["QB"],
        "BLOCK_SN2": lambda args: args["N2"] // args["QB"],
    }
)
@triton.jit
def _rms_norm_fwd_fused(
    X,  # pointer to the input
    SX,  # pointer to the scale of input
    Y,  # pointer to the output
    SY,  # pointer to the scale of output
    W,  # Weight
    Rstd,  # pointer to the 1/std
    stride,  # how much to increase the pointer when moving by 1 row
    scale_stride,  # how much to increase the pointer when moving by 1 row
    N: tl.constexpr,  # number of columns in X,
    N2: tl.constexpr,  # number of columns in X,
    SN: tl.constexpr,
    SN2: tl.constexpr,
    QB: tl.constexpr,
    eps,  # epsilon to avoid division by zero
    fp8_max,
    SCALE_MIN_THRES: tl.constexpr,
    BLOCK_SN: tl.constexpr,
    BLOCK_SN2: tl.constexpr,
):
    # Map the program id to the row of X and Y it should compute.
    row = tl.program_id(0)
    Y += row * stride
    X += row * stride
    SX += row * scale_stride

    cols = tl.arange(0, N2)
    scale_cols = tl.arange(0, SN2)
    x = tl.load(X + cols, mask=cols < N, other=0.0).to(tl.float32)
    scale_x = tl.load(SX + scale_cols, mask=scale_cols < SN, other=0.0).to(tl.float32)

    # Dequantize and swish calculation
    scale_x = tl.reshape(scale_x, (BLOCK_SN2, 1))
    x = tl.reshape(x, (BLOCK_SN2, QB))
    x = x * scale_x
    x = tl.reshape(x, N2)

    # Compute variance
    _var = x * x
    var = tl.sum(_var, axis=0) / N
    rstd = 1 / tl.sqrt(var + eps)

    # Write mean / rstd
    tl.store(Rstd + row, rstd)

    # Normalize and apply linear transformation
    w = tl.load(W + cols)
    x_hat = x * rstd
    y = x_hat * w

    # Scale calculation
    abs_y = tl.abs(y)
    max_val = tl.max(abs_y) + SCALE_MIN_THRES
    scale_output = max_val / fp8_max
    tl.store(SY + row, scale_output)

    # Write output
    tl.store(Y + cols, y, mask=cols < N)


@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["N"] // args["QB"],
        "BLOCK_SN2": lambda args: args["N2"] // args["QB"],
    }
)
@triton.jit
def _rms_norm_bwd_dx_fused(
    DX,  # pointer to the input gradient
    DY,  # pointer to the output gradient
    DW,
    X,  # pointer to the input
    SX,  # pointer to the input
    W,  # weight
    Rstd,  # pointer to the 1/std
    Lock,
    stride,  # how much to increase the pointer when moving by 1 row
    scale_stride,  # how much to increase the pointer when moving by 1 row
    N: tl.constexpr,  # number of columns in X,
    N2: tl.constexpr,  # number of columns in X,
    SN: tl.constexpr,
    SN2: tl.constexpr,
    QB: tl.constexpr,
    SCALE_MIN_THRES: tl.constexpr,
    BLOCK_SN: tl.constexpr,
    BLOCK_SN2: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    # Map the program id to the elements of X, DX, and DY it should compute.
    row = tl.program_id(0)
    cols = tl.arange(0, N2)
    scale_cols = tl.arange(0, SN2)
    mask = cols < N
    X += row * stride
    DY += row * stride
    DX += row * stride
    SX += row * scale_stride

    # Offset locks and weights/biases gradient pointer for parallel reduction
    lock_id = row % GROUP_SIZE_M
    Lock += lock_id
    Count = Lock + GROUP_SIZE_M
    DW = DW + lock_id * N + cols

    # Load data to SRAM
    x = tl.load(X + cols, mask=mask, other=0.0).to(tl.float32)
    scale_x = tl.load(SX + scale_cols, mask=scale_cols < SN, other=0.0).to(tl.float32)
    scale_x = tl.reshape(scale_x, (BLOCK_SN2, 1))
    x = tl.reshape(x, (BLOCK_SN2, QB))
    x = x * scale_x
    x = tl.reshape(x, N2)
    dy = tl.load(DY + cols, mask=mask, other=0.0).to(tl.float32)

    # Load weight
    w = tl.load(W + cols).to(tl.float32)
    rstd = tl.load(Rstd + row).to(tl.float32)

    # Compute dx
    xhat = x * rstd
    wdy = w * dy
    xhat = tl.where(mask, xhat, 0.0)
    wdy = tl.where(mask, wdy, 0.0)
    c1 = tl.sum(xhat * wdy, axis=0) / N
    dx = (wdy - (xhat * c1)) * rstd  # layer norm have c2 term, rmsnorm do not

    dx = dx.to(DX.type.element_ty)

    # Write dx
    tl.store(DX + cols, dx, mask=mask)

    # Accumulate partial sums for dw/db
    partial_dw = (dy * xhat).to(w.dtype)
    while tl.atomic_cas(Lock, 0, 1) == 1:
        pass
    count = tl.load(Count)
    # First store doesn't accumulate
    if count == 0:
        tl.atomic_xchg(Count, 1)
    else:
        partial_dw += tl.load(DW, mask=mask)
    tl.store(DW, partial_dw, mask=mask)
    # Release the lock
    tl.atomic_xchg(Lock, 0)


@triton.jit
def _rms_norm_bwd_dwdb(
    DW,  # pointer to the partial sum of weights gradient
    FINAL_DW,  # pointer to the weights gradient
    M,  # GROUP_SIZE_M
    N,  # number of columns
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
):
    # Map the program id to the elements of DW and DB it should compute.
    pid = tl.program_id(0)
    cols = pid * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    dw = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    # Iterate through the rows of DW and DB to sum the partial sums.
    for i in range(0, M, BLOCK_SIZE_M):
        rows = i + tl.arange(0, BLOCK_SIZE_M)
        mask = (rows[:, None] < M) & (cols[None, :] < N)
        offs = rows[:, None] * N + cols[None, :]
        dw += tl.load(DW + offs, mask=mask, other=0.0)
    # Write the final sum to the output.
    sum_dw = tl.sum(dw, axis=0)
    tl.store(FINAL_DW + cols, sum_dw, mask=cols < N)


def fp8_rmsnorm_forward(x, s_x, w, QB, eps, transpose_output_2d=False):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        assert len(s_x.shape) == 3
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])
        s_x = s_x.reshape(-1, s_x.shape[-1])

    # allocate output
    M, N = x.shape
    _, SN = s_x.shape
    y = torch.empty_like(x, dtype=torch.bfloat16)
    s_y = torch.empty((M,), dtype=torch.bfloat16, device=x.device)
    rstd = torch.empty((M,), dtype=torch.float32, device=x.device)
    # heuristics for number of warps
    num_warps = 8
    fp8MaxValue = FP8_MAX_VALUE[x.dtype]

    N2 = triton.next_power_of_2(N)
    SN2 = N2 // QB
    # enqueue kernel
    _rms_norm_fwd_fused[(M,)](  #
        x,
        s_x,
        y,
        s_y,
        w,
        rstd,  #
        x.stride(0),
        s_x.stride(0),
        N,
        N2,
        SN,
        SN2,
        QB,
        eps,
        fp8MaxValue,
        SCALE_MIN_THRES=SCALE_MIN_THRES,  #
        num_warps=num_warps,
        num_ctas=1,
    )

    # reduction
    s_y_max = s_y.max()
    qy, s_y_max, qy_t = fp8_division_transpose(y, QB, x.dtype, s_y_max)

    # Recover 2D to 3D
    if batched:
        y = y.reshape(BS, -1, y.shape[-1])
        qy = qy.reshape(BS, -1, y.shape[-1])
        if not transpose_output_2d:
            qy_t = qy_t.reshape(BS, -1, y.shape[-1])

    return qy, s_y_max, qy_t, (w, rstd, num_warps)


def fp8_rmsnorm_backward(x, s_x, g, w, v, QB, num_warps):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        assert len(s_x.shape) == 3
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])
        s_x = s_x.reshape(-1, s_x.shape[-1])
        g = g.reshape(-1, g.shape[-1])

    # enqueue kernel using forward pass heuristics
    # also compute partial sums for DW and DB
    M, N = g.shape
    _, SN = s_x.shape

    GROUP_SIZE_M = 128
    # heuristics for amount of parallel reduction stream for DW/DB
    locks = torch.zeros(2 * GROUP_SIZE_M, dtype=torch.int32, device=w.device)
    _dw = torch.zeros((GROUP_SIZE_M, N), dtype=w.dtype, device=w.device)
    dw = torch.empty((N,), dtype=w.dtype, device=w.device)

    dx = torch.empty_like(g, dtype=torch.bfloat16)

    N2 = triton.next_power_of_2(N)
    SN2 = triton.next_power_of_2(SN)
    _rms_norm_bwd_dx_fused[(M,)](  #
        dx,
        g,
        _dw,
        x,
        s_x,
        w,
        v,
        locks,  #
        x.stride(0),
        s_x.stride(0),
        N,
        N2,
        SN,
        SN2,
        QB,
        SCALE_MIN_THRES=SCALE_MIN_THRES,
        num_warps=num_warps,
        GROUP_SIZE_M=GROUP_SIZE_M,
    )

    if batched:
        dx = dx.reshape(BS, -1, dx.shape[-1])

    grid = lambda meta: [triton.cdiv(N, meta["BLOCK_SIZE_N"])]
    # accumulate partial sums in separate kernel
    _rms_norm_bwd_dwdb[grid](_dw, dw, min(GROUP_SIZE_M, M), N, BLOCK_SIZE_M=32, BLOCK_SIZE_N=128, num_ctas=1)  #  #

    return dx, dw



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
# 4 block
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice

from .common import (FP8_MAX_VALUE, SCALE_MIN_THRES, convert_fp8_to_embit,
                     convert_str_to_fp8, get_configs_io_block)

"""Quantize and Transpose Operator"""
"""Input uses 1 * 16 group quantization"""
"""Output uses 1 * 16 group quantization"""
"""The input can be 2D or 3D, but the calculation is performed in 2D"""


@triton.autotune(
    configs=[] + get_configs_io_block(),
    key=[
        "M",
        "N",
    ],
)
@triton.heuristics(
    {
        "BLOCK_SN": lambda args: args["BLOCK_N"] // args["QB"],
    }
)
@triton.jit
def _fp8_division_kernel(
    output_ptr,  # output
    input_ptr,
    input_scale_ptr,  # input
    noise_ptr,  # noise for stochastic
    M,
    N,
    SN,
    QB: tl.constexpr,
    fp8_max,
    e_bit: tl.constexpr,
    m_bit: tl.constexpr,  # shape
    input_stride_0,
    input_stride_1,  # input stride
    output_stride_0,
    output_stride_1,  # output stride
    SCALE_MIN_THRES: tl.constexpr,  # We do not use it since we believe SCALE_MIN_THRES should be used in previous kernel when calculating scaling factor
    STOCHASTIC: tl.constexpr,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_SN: tl.constexpr,
):  # CUDA block size

    # Block PID
    pid = tl.program_id(0)
    NUM_BLOCK_N = tl.cdiv(N, BLOCK_N)
    pid_dim0 = pid // NUM_BLOCK_N
    pid_dim1 = pid % NUM_BLOCK_N

    # pointers
    input_block_ptr = tl.make_block_ptr(
        base=input_ptr,
        shape=(M, N),
        strides=(input_stride_0, input_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    input = tl.load(input_block_ptr)
    input = input.to(tl.float32)
    scale_output = tl.load(input_scale_ptr)
    scale_output = scale_output.to(tl.float32)

    output = tl.reshape(input, (BLOCK_M, BLOCK_SN, QB))

    # Quantize Scale calculation
    # Quantize
    output = tl.fdiv(output, scale_output)
    output = tl.reshape(output, (BLOCK_M, BLOCK_N))

    if STOCHASTIC:
        # noise_block_ptr = tl.make_block_ptr(
        #     base=noise_ptr,
        #     shape=(M, N),
        #     strides=(input_stride_0, input_stride_1),
        #     offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        #     block_shape=(BLOCK_M, BLOCK_N),
        #     order=(1, 0)
        # )
        # noise = tl.load(noise_block_ptr)

        offs_m = pid_dim0 * BLOCK_M + tl.arange(0, BLOCK_M)
        offs_n = pid_dim1 * BLOCK_N + tl.arange(0, BLOCK_N)
        noise_offset = offs_m[:, None] * input_stride_0 + offs_n[None, :] * input_stride_1
        noise = tl.rand(0, noise_offset)

        output = _stochastic_rounding(output, noise, e_bit, m_bit)

    output = output.to(output_ptr.type.element_ty)

    # pointers
    output_block_ptr = tl.make_block_ptr(
        base=output_ptr,
        shape=(M, N),
        strides=(output_stride_0, output_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    tl.store(output_block_ptr, output)


@triton.jit
def _stochastic_rounding(output, noise, e_bit: tl.constexpr, m_bit: tl.constexpr):
    subnormal_min = tl.exp2(2 - tl.exp2(e_bit - 1) - m_bit)
    # subnormal_should_be = tl.exp2(2 - tl.exp2(e_bit) - 1)

    output_int32 = tl.cast(output, tl.int32, bitcast=True)
    output_int32 = output_int32 & 0x7F800000
    output_float32 = tl.cast(output_int32, tl.float32, bitcast=True)
    output_exp = tl.maximum(output_float32, subnormal_min)

    noise_rescale = tl.exp2(m_bit) + (output_exp == subnormal_min) * (
        1 - tl.exp2(m_bit)
    )  # 2^m_bit for normal, 1 for subnormal

    noise = output_exp * noise / noise_rescale
    sign = 1 - 2 * libdevice.signbit(output)
    output = tl.abs(output) + noise

    minmax_ratio = 2 + (output_exp == subnormal_min) * (tl.exp2(m_bit) - 2)  # 2 for normal, and 2^M for subnormal
    output = sign * tl.clamp(output, min=output_exp, max=minmax_ratio * output_exp)

    return output


def fp8_division(x, QB, fp8type, s_y=None, stochastic=False):
    # Change batched 3D input to 2D
    batched = False
    if len(x.shape) == 3:
        batched = True
        BS = x.shape[0]
        x = x.reshape(-1, x.shape[-1])

    if stochastic:
        # noise = torch.zeros_like(x, dtype=torch.float32).uniform_(-0.5, 0.5)
        noise = None
    else:
        noise = None

    # defining the input and output tensor
    M, N = x.shape
    SN = N // QB

    if isinstance(fp8type, str):
        fp8type = convert_str_to_fp8[fp8type]

    y = torch.empty_like(x, dtype=fp8type)
    fp8MaxValue = FP8_MAX_VALUE[fp8type]  # E4M3 and E5M2 have different max value
    e_bit, m_bit = convert_fp8_to_embit[fp8type]

    if s_y is None:
        s_y = (x.abs().max() + SCALE_MIN_THRES) / fp8MaxValue

    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)

    _fp8_division_kernel[grid](
        y,
        x,
        s_y,
        noise,
        M,
        N,
        SN,
        QB,
        fp8MaxValue,
        e_bit,
        m_bit,
        x.stride(0),
        x.stride(1),
        y.stride(0),
        y.stride(1),
        SCALE_MIN_THRES=SCALE_MIN_THRES,
        STOCHASTIC=stochastic,
    )

    # Recover 2D to 3D
    if batched:
        y = y.reshape(BS, -1, y.shape[-1])

    return y, s_y  # y_t is expected to be 2D tensor



from dataclasses import dataclass

from transformers import PretrainedConfig


@dataclass
class QuantizationConfig:
    # Quantize Activation
    quantize_model: str = "true"
    symm: bool = True
    epsilon: float = 1e-10
    fabit: str = "E4M3"
    fwbit: str = "E4M3"
    fobit: str = "E4M3"
    babit: str = "E5M2"
    bwbit: str = "E5M2"
    bobit: str = "E5M2"
    group_size: int = 16
    pad_to_multiple_of: int = 8
    weight_memory_efficient: bool = True

    # Quantize Optimizer States
    first_order_expansion: str = "true"
    second_order_expansion: str = "true"
    first_order_bit: str = "E4M3"
    second_order_bit: str = "E4M3"
    qgroup_size: int = 128
    expand_min: int = 16

    # Legacy
    row_blocksize: int = -1
    col_blocksize: int = -1

    def __init__(
        self,
        quantize_model: str = "true",
        symm: bool = True,
        epsilon: float = 1e-10,
        fabit: str = "E4M3",
        fwbit: str = "E4M3",
        fobit: str = "E4M3",
        babit: str = "E5M2",
        bwbit: str = "E5M2",
        bobit: str = "E5M2",
        group_size: int = 16,
        pad_to_multiple_of: int = 0,
        weight_memory_efficient: bool = True,
        first_order_expansion: str = "true",
        second_order_expansion: str = "true",
        first_order_bit: str = "E4M3",
        second_order_bit: str = "E4M3",
        qgroup_size: int = 128,
        expand_min: int = 16,
        row_blocksize: int = -1,
        col_blocksize: int = -1,
        **kwargs,
    ):
        super().__init__()
        self.quantize_model = quantize_model
        self.symm = symm
        self.epsilon = epsilon
        self.fabit = fabit
        self.fwbit = fwbit
        self.fobit = fobit
        self.babit = babit
        self.bwbit = bwbit
        self.bobit = bobit
        self.group_size = group_size
        self.pad_to_multiple_of = pad_to_multiple_of
        self.weight_memory_efficient = weight_memory_efficient

        self.first_order_expansion = first_order_expansion
        self.second_order_expansion = second_order_expansion
        self.first_order_bit = first_order_bit
        self.second_order_bit = second_order_bit
        self.qgroup_size = qgroup_size
        self.expand_min = expand_min
        
        self.row_blocksize = row_blocksize
        self.col_blocksize = col_blocksize


# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

from collections import OrderedDict, defaultdict
from copy import deepcopy
from itertools import chain
from typing import (Any, DefaultDict, Dict, Hashable, Iterable, List, Optional,
                    Tuple, Union)

import qoptim_cuda
import torch
from torch import Tensor
from torch.optim.optimizer import Optimizer
from typing_extensions import ParamSpec, Self, TypeAlias

from ..utils._fp8_quantization_config import QuantizationConfig

StateDict: TypeAlias = Dict[str, Any]

convert_str_to_fp8 = {"E4M3": torch.float8_e4m3fn, "E5M2": torch.float8_e5m2}


class CoatAdamW(Optimizer):
    def __init__(
        self,
        params,
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-8,
        weight_decay: float = 1e-2,
        amsgrad: bool = False,
        qargs: QuantizationConfig = None,
        *,
        fused: Optional[bool] = None,
    ):
        if qargs is None:
            qargs = QuantizationConfig()
        self.qargs = qargs

        assert self.qargs.first_order_expansion == self.qargs.second_order_expansion, "We expect the first order and second order momentum to use dynamic range expansion at the same time."
        if not 0.0 <= lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if not 0.0 <= eps:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")
        if not 0.0 <= weight_decay:
            raise ValueError(f"Invalid weight_decay value: {weight_decay}")
        defaults = dict(
            lr=lr,
            betas=betas,
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            fused=fused,
        )
        super().__init__(params, defaults)

    def __setstate__(self, state):
        super().__setstate__(state)
        for group in self.param_groups:
            group.setdefault("amsgrad", False)
            fused = group.setdefault("fused", None)
            for p in group["params"]:
                p_state = self.state.get(p, [])
                if len(p_state) != 0 and not torch.is_tensor(p_state["step"]):
                    step_val = float(p_state["step"])
                    p_state["step"] = torch.tensor(step_val, dtype=torch.float32)

    def _init_group(
        self,
        group,
        params_with_grad,
        grads,
        amsgrad,
        use_expansion,
        exp_avgs,
        scale_exp_avgs,
        expand_exp_avgs,
        sqrt_minmax_exp_avgs,
        exp_avg_sqs,
        scale_exp_avg_sqs,
        expand_exp_avg_sqs,
        sqrt_minmax_exp_avg_sqs,
        max_exp_avg_sqs,
        state_steps,
    ):
        for p in group["params"]:
            if p.grad is None:
                continue
            params_with_grad.append(p)
            if p.grad.is_sparse:
                raise RuntimeError("AdamW does not support sparse gradients")
            grads.append(p.grad)

            state = self.state[p]

            # print(f'Param shape: {p.shape}', file=open('debug.txt', 'a'))
            # print(f'Param shape: {p.shape}, {p.device}')

            # State initialization
            if len(state) == 0:
                # This is because kernel launches are costly on CUDA and XLA.
                state["step"] = torch.tensor(0.0)

                # Should be torch.float8_e4m3fn
                first_order_dtype = convert_str_to_fp8[self.qargs.first_order_bit]
                second_order_dtype = convert_str_to_fp8[self.qargs.second_order_bit]
                scale_shape = (p.numel() + self.qargs.qgroup_size - 1) // self.qargs.qgroup_size

                # Exponential moving average of gradient values
                state["exp_avg"] = torch.zeros_like(p, dtype=first_order_dtype, memory_format=torch.preserve_format)
                state["scale_exp_avg"] = torch.zeros(scale_shape, device=p.device, dtype=p.dtype)
                if use_expansion:
                    state["expand_exp_avg"] = torch.ones(scale_shape, device=p.device, dtype=p.dtype)
                    state["sqrt_minmax_exp_avg"] = torch.ones(scale_shape, device=p.device, dtype=p.dtype)
                # Exponential moving average of squared gradient values
                state["exp_avg_sq"] = torch.zeros_like(p, dtype=second_order_dtype, memory_format=torch.preserve_format)
                state["scale_exp_avg_sq"] = torch.zeros(scale_shape, device=p.device, dtype=p.dtype)
                if use_expansion:
                    state["expand_exp_avg_sq"] = torch.ones(scale_shape, device=p.device, dtype=p.dtype)
                    state["sqrt_minmax_exp_avg_sq"] = torch.ones(scale_shape, device=p.device, dtype=p.dtype)
                if amsgrad:
                    # Maintains max of all exp. moving avg. of sq. grad. values
                    state["max_exp_avg_sq"] = torch.zeros(p, memory_format=torch.preserve_format)

            exp_avgs.append(state["exp_avg"])
            scale_exp_avgs.append(state["scale_exp_avg"])
            if use_expansion:
                expand_exp_avgs.append(state["expand_exp_avg"])
                sqrt_minmax_exp_avgs.append(state["sqrt_minmax_exp_avg"])
            exp_avg_sqs.append(state["exp_avg_sq"])
            scale_exp_avg_sqs.append(state["scale_exp_avg_sq"])
            if use_expansion:
                expand_exp_avg_sqs.append(state["expand_exp_avg_sq"])
                sqrt_minmax_exp_avg_sqs.append(state["sqrt_minmax_exp_avg_sq"])

            if group["amsgrad"]:
                max_exp_avg_sqs.append(state["max_exp_avg_sq"])

            state_steps.append(state["step"])

    @torch._disable_dynamo
    def load_state_dict(self, state_dict: StateDict) -> None:
        r"""Loads the optimizer state.

        Args:
            state_dict (dict): optimizer state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        # shallow copy, to be consistent with module API
        state_dict = state_dict.copy()

        for pre_hook in self._optimizer_load_state_dict_pre_hooks.values():
            hook_result = pre_hook(self, state_dict)
            if hook_result is not None:
                state_dict = hook_result

        # Validate the state_dict
        groups = self.param_groups

        # Deepcopy as we write into saved_groups later to update state
        saved_groups = deepcopy(state_dict["param_groups"])

        if len(groups) != len(saved_groups):
            raise ValueError("loaded state dict has a different number of " "parameter groups")
        param_lens = (len(g["params"]) for g in groups)
        saved_lens = (len(g["params"]) for g in saved_groups)
        if any(p_len != s_len for p_len, s_len in zip(param_lens, saved_lens)):
            raise ValueError(
                "loaded state dict contains a parameter group " "that doesn't match the size of optimizer's group"
            )

        # Update the state
        id_map = dict(
            zip(
                chain.from_iterable(g["params"] for g in saved_groups), chain.from_iterable(g["params"] for g in groups)
            )
        )

        def _cast(param, value, param_id=None, param_groups=None, key=None):
            r"""Make a deep copy of value, casting all tensors to device of param."""
            if isinstance(value, torch.Tensor):
                return CoatAdamW._process_value_according_to_param_policy(param, value, param_id, param_groups, key)
            elif isinstance(value, dict):
                return {
                    k: _cast(param, v, param_id=param_id, param_groups=param_groups, key=k) for k, v in value.items()
                }
            elif isinstance(value, Iterable):
                return type(value)(_cast(param, v, param_id=param_id, param_groups=param_groups) for v in value)  # type: ignore[call-arg]
            else:
                return value

        # Copy state assigned to params (and cast tensors to appropriate types).
        # State that is not assigned to params is copied as is (needed for
        # backward compatibility).
        state: DefaultDict[torch.Tensor, Dict[Any, Any]] = defaultdict(dict)
        for k, v in state_dict["state"].items():
            if k in id_map:
                param = id_map[k]
                state[param] = _cast(param, v, param_id=k, param_groups=state_dict["param_groups"])
            else:
                state[k] = v

        # Update parameter groups, setting their 'params' value
        def update_group(group: Dict[str, Any], new_group: Dict[str, Any]) -> Dict[str, Any]:
            new_group["params"] = group["params"]
            return new_group

        param_groups = [update_group(g, ng) for g, ng in zip(groups, saved_groups)]
        self.__setstate__({"state": state, "param_groups": param_groups})

        for post_hook in self._optimizer_load_state_dict_post_hooks.values():
            post_hook(self)

    @staticmethod
    def _process_value_according_to_param_policy(
        param: torch.Tensor,
        value: torch.Tensor,
        param_id: int,
        param_groups: List[Dict[Any, Any]],
        key: Hashable = None,
    ) -> torch.Tensor:
        # Floating-point types are a bit special here. They are the only ones
        # that are assumed to always match the type of params.
        # Make sure state['step'] is not casted https://github.com/pytorch/pytorch/issues/74424
        # UNLESS fused or capturable, see note [special device hosting for step]
        fused = False
        capturable = False
        assert param_groups is not None
        for pg in param_groups:
            if param_id in pg["params"]:
                fused = pg["fused"] if "fused" in pg else False
                capturable = pg["capturable"] if "capturable" in pg else False
                break
        if key == "step":
            if capturable or fused:
                return value.to(dtype=torch.float32, device=param.device)
            else:
                return value
        else:
            assert value.dtype in [torch.float8_e4m3fn, torch.float8_e5m2, torch.float32]
            return value.to(device=param.device)  # do not cast optimizer states
            # if param.is_floating_point():
            #     return value.to(dtype=param.dtype, device=param.device)
            # else:
            #     return value.to(device=param.device)

    @torch.no_grad()
    def step(self, closure=None):
        """Perform a single optimization step.

        Args:
            closure (Callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        self._cuda_graph_capture_health_check()

        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            params_with_grad = []
            grads = []
            exp_avgs = []
            scale_exp_avgs = []
            expand_exp_avgs = []
            sqrt_minmax_exp_avgs = []
            exp_avg_sqs = []
            scale_exp_avg_sqs = []
            expand_exp_avg_sqs = []
            sqrt_minmax_exp_avg_sqs = []
            max_exp_avg_sqs = []
            state_steps = []
            amsgrad = group["amsgrad"]
            use_expansion = self.qargs.first_order_expansion.lower() in ["expansion", "true"]
            beta1, beta2 = group["betas"]

            self._init_group(
                group,
                params_with_grad,
                grads,
                amsgrad,
                use_expansion,
                exp_avgs,
                scale_exp_avgs,
                expand_exp_avgs,
                sqrt_minmax_exp_avgs,
                exp_avg_sqs,
                scale_exp_avg_sqs,
                expand_exp_avg_sqs,
                sqrt_minmax_exp_avg_sqs,
                max_exp_avg_sqs,
                state_steps,
            )

            Coatadamw(
                self.qargs,
                params_with_grad,
                grads,
                exp_avgs,
                scale_exp_avgs,
                expand_exp_avgs,
                sqrt_minmax_exp_avgs,
                exp_avg_sqs,
                scale_exp_avg_sqs,
                expand_exp_avg_sqs,
                sqrt_minmax_exp_avg_sqs,
                max_exp_avg_sqs,
                state_steps,
                amsgrad=amsgrad,
                use_expansion=use_expansion,
                beta1=beta1,
                beta2=beta2,
                lr=group["lr"],
                weight_decay=group["weight_decay"],
                eps=group["eps"],
                qgroup_size=self.qargs.qgroup_size,
                expand_min=self.qargs.expand_min,
                fused=group["fused"],
                grad_scale=getattr(self, "grad_scale", None),
                found_inf=getattr(self, "found_inf", None),
            )

        return loss


def Coatadamw(
    qargs,
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    scale_exp_avgs: List[Tensor],
    expand_exp_avgs: List[Tensor],
    sqrt_minmax_exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    scale_exp_avg_sqs: List[Tensor],
    expand_exp_avg_sqs: List[Tensor],
    sqrt_minmax_exp_avg_sqs: List[Tensor],
    max_exp_avg_sqs: List[Tensor],
    state_steps: List[Tensor],
    # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
    # setting this as kwarg for now as functional API is compiled by torch/distributed/optim
    fused: Optional[bool] = None,
    grad_scale: Optional[Tensor] = None,
    found_inf: Optional[Tensor] = None,
    *,
    amsgrad: bool,
    use_expansion: bool,
    beta1: float,
    beta2: float,
    lr: Union[float, Tensor],
    weight_decay: float,
    eps: float,
    qgroup_size: int,
    expand_min: int,
):
    r"""Functional API that performs AdamW algorithm computation.

    See :class:`~torch.optim.AdamW` for details.
    """
    if not torch._utils.is_compiling() and not all(isinstance(t, torch.Tensor) for t in state_steps):
        raise RuntimeError("API has changed, `state_steps` argument must contain a list of singleton tensors")

    func = _single_tensor_Coatadamw

    func(
        qargs,
        params,
        grads,
        exp_avgs,
        scale_exp_avgs,
        expand_exp_avgs,
        sqrt_minmax_exp_avgs,
        exp_avg_sqs,
        scale_exp_avg_sqs,
        expand_exp_avg_sqs,
        sqrt_minmax_exp_avg_sqs,
        max_exp_avg_sqs,
        state_steps,
        amsgrad=amsgrad,
        use_expansion=use_expansion,
        beta1=beta1,
        beta2=beta2,
        lr=lr,
        weight_decay=weight_decay,
        eps=eps,
        qgroup_size=qgroup_size,
        expand_min=expand_min,
        grad_scale=grad_scale,
        found_inf=found_inf,
    )


def _dispatch_sqrt(x: float):  # float annotation is needed because of torchscript type inference
    if not torch.jit.is_scripting() and isinstance(x, torch.Tensor):
        return x.sqrt()
    else:
        return sqrt(x)


def _single_tensor_Coatadamw(
    qargs,
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    scale_exp_avgs: List[Tensor],
    expand_exp_avgs: List[Tensor],
    sqrt_minmax_exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    scale_exp_avg_sqs: List[Tensor],
    expand_exp_avg_sqs: List[Tensor],
    sqrt_minmax_exp_avg_sqs: List[Tensor],
    max_exp_avg_sqs: List[Tensor],
    state_steps: List[Tensor],
    grad_scale: Optional[Tensor],
    found_inf: Optional[Tensor],
    *,
    amsgrad: bool,
    use_expansion: bool,
    beta1: float,
    beta2: float,
    lr: Union[Tensor, float],
    weight_decay: float,
    eps: float,
    qgroup_size: int,
    expand_min: int,
):

    assert grad_scale is None and found_inf is None

    if torch.jit.is_scripting():
        # this assert is due to JIT being dumb and not realizing that the ops below
        # have overloads to handle both float and Tensor lrs, so we just assert it's
        # a float since most people using JIT are using floats
        assert isinstance(lr, float)

    for i, param in enumerate(params):
        grad = grads[i]
        # First order
        exp_avg = exp_avgs[i]
        scale_exp_avg = scale_exp_avgs[i]
        # Second order
        exp_avg_sq = exp_avg_sqs[i]
        scale_exp_avg_sq = scale_exp_avg_sqs[i]
        step_t = state_steps[i]

        # print(len(exp_avg.unique()), len(exp_avg_sq.unique()))
        # print(f"{param.shape}, {grad.shape}, {exp_avg.shape}, {exp_avg_sq.shape}", file=open('debug.txt', 'a'))

        # update step
        step_t += 1
        step = int(step_t.item())

        # Perform Optimizer Step
        if use_expansion:
            expand_exp_avg = expand_exp_avgs[i]
            sqrt_minmax_exp_avg = sqrt_minmax_exp_avgs[i]
            expand_exp_avg_sq = expand_exp_avg_sqs[i]
            sqrt_minmax_exp_avg_sq = sqrt_minmax_exp_avg_sqs[i]

            qoptim_cuda.fp8_adamw_expand_step(
                param,
                grad,
                exp_avg,
                scale_exp_avg,
                expand_exp_avg,
                sqrt_minmax_exp_avg,
                exp_avg_sq,
                scale_exp_avg_sq,
                expand_exp_avg_sq,
                sqrt_minmax_exp_avg_sq,
                beta1,
                beta2,
                lr,
                weight_decay,
                eps,
                step,
                qgroup_size,
                expand_min,
            )

        else:
            qoptim_cuda.fp8_adamw_step(
                param,
                grad,
                exp_avg,
                scale_exp_avg,
                exp_avg_sq,
                scale_exp_avg_sq,
                beta1,
                beta2,
                lr,
                weight_decay,
                eps,
                step,
                qgroup_size,
            )



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import torch
# 4 block
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice

CONST_BLOCK = 32

# The kernel with 1 load operation and 4 store operation
def get_configs_io_block():
    configs = []
    for nstages in [3, 4, 5, 6]:
        for block_m in [32, 64, 128]:
            for block_n in [32, 64, 128]:
                for nwarps in [4, 8, 16, 32]:
                    configs.append(
                        triton.Config(
                            {"BLOCK_M": block_m, "BLOCK_N": block_n},
                            num_stages=nstages,
                            num_warps=nwarps,
                        )
                    )
    return configs


@triton.autotune(
    configs=[] + get_configs_io_block(),
    key=[
        "M",
        "N",
    ],
)
@triton.jit
def bench_memory_io_kernel_forward(
    output_ptr,
    input_ptr,
    M,
    N,
    B: tl.constexpr,
    input_stride_0,
    input_stride_1,
    output_stride_0,
    output_stride_1,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):

    # Block PID
    pid = tl.program_id(0)
    NUM_BLOCK_M = tl.cdiv(M, BLOCK_M)
    NUM_BLOCK_N = tl.cdiv(N, BLOCK_N)
    pid_dim0 = pid // NUM_BLOCK_N
    pid_dim1 = pid % NUM_BLOCK_N

    # pointers
    input_block_ptr = tl.make_block_ptr(
        base=input_ptr,
        shape=(M, N),
        strides=(input_stride_0, input_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    input = tl.load(input_block_ptr)
    input = input.to(tl.float32)

    output = input * 2

    # pointers
    output_block_ptr = tl.make_block_ptr(
        base=output_ptr,
        shape=(M, N),
        strides=(output_stride_0, output_stride_1),
        offsets=(pid_dim0 * BLOCK_M, pid_dim1 * BLOCK_N),
        block_shape=(BLOCK_M, BLOCK_N),
        order=(1, 0),
    )

    output = output.to(output_ptr.type.element_ty)
    tl.store(output_block_ptr, output)


def bench_memory_io_forward(x, B):
    # defining the input and output tensor
    M, N = x.shape

    y = torch.empty_like(x, dtype=x.dtype)

    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)

    bench_memory_io_kernel_forward[grid](
        y,
        x,
        M,
        N,
        B,
        x.stride(0),
        x.stride(1),
        y.stride(0),
        y.stride(1),
    )
    return y


configs = []
for SL in [8192]:
    configs.append(
        triton.testing.Benchmark(  # test different matrix size influence
            x_names=["CDIM"],
            x_vals=[1024, 2048, 4096, 8192],
            line_arg="dtype",
            line_vals=[torch.int8, torch.float16, torch.float32],
            line_names=["float8", "float16", "float32"],
            styles=[("blue", "-"), ("green", "-"), ("red", "-")],
            ylabel="time-cost",
            plot_name=f"INT8GELU<BLSZ={CONST_BLOCK}><SL={SL}>",
            args={"SL": SL, "B": CONST_BLOCK, "provider": "triton", "mode": "time-consuming"},
        )
    )


@triton.testing.perf_report(configs)
def bench_load_store(
    SL, CDIM, B, provider, dtype, mode="forward"
):  # I only use triton as the provider, and mode when benchmarking
    # create data
    x = torch.randn(SL, CDIM, dtype=torch.float32).cuda()
    x = x.to(dtype)

    quantiles = [0.5, 0.2, 0.8]
    # utility functions
    if provider == "triton":

        def y_fwd():
            bench_memory_io_forward(x, B)

    if provider == "torch":
        torch_gelu = torch.nn.GELU()

        def y_fwd():
            return torch_gelu(x)

    # forward pass
    if mode == "time-consuming":
        convert_func = lambda ms: ms
        ms, min_ms, max_ms = triton.testing.do_bench(y_fwd, quantiles=quantiles, rep=100)
    # backward pass
    if mode == "gbps":
        convert_func = lambda ms: 2 * x.numel() * x.element_size() / ms * 1e-6
        ms, min_ms, max_ms = triton.testing.do_bench(y_fwd, quantiles=quantiles, rep=100)
    return convert_func(ms), convert_func(max_ms), convert_func(min_ms)


if __name__ == "__main__":
    torch.manual_seed(0)
    torch.set_printoptions(precision=8, linewidth=1600, sci_mode=False, edgeitems=3)
    bench_load_store.run(print_data=True)



# Activation
# Utils
from ._dequantize import fp8_dequantize
from ._division import fp8_division
from ._division_transpose import fp8_division_transpose
from ._quantize import fp8_quantize
from ._quantize_pertensor import fp8_quantize_pertensor
from ._quantize_pertensor_transpose import fp8_quantize_pertensor_transpose
from ._quantize_perblock import fp8_quantize_perblock
from ._transpose import fp8_transpose
from .add_bwd import fp8_add_Ifp_Ifp_Ofp_Opt
from .add_fwd import fp8_add_Ifp_Ifp_Ofp_Og16
# Normalization
from .func_layernorm_noparam import fp8_layernorm_noparam_backward, fp8_layernorm_noparam_forward
from .func_layernorm_param import fp8_layernorm_param_backward, fp8_layernorm_param_forward
from .func_quantize import Coat_quantize_bgn, Coat_quantize_end
from .func_rmsnorm import fp8_rmsnorm_backward, fp8_rmsnorm_forward
from .gelu_bwd import fp8_gelu_backward
from .gelu_fwd import fp8_gelu_forward
# linear and add
from .linear import fp8_linear_backward, fp8_linear_forward
from .mul_bwd import fp8_mul_backward
from .mul_fwd import fp8_mul_forward
from .silu_bwd import fp8_silu_backward
from .silu_fwd import fp8_silu_forward



import torch.nn as nn
from ..activation.real_quantization import fp8_division_transpose

class FP8CacheWeightModule(nn.Module):
    def __init__(self, config, qargs, layer_id):
        super().__init__()
        self.config = config
        self.qargs = qargs
        self.layer_id = layer_id

    def prepare_weight(self, weight, weight_name, is_first_microbatch):
        if is_first_microbatch:
            if self.qargs.weight_memory_efficient:
                # print(f"{weight_name} uses first microbatch")
                weight_fp8, weight_s, weight_fp8_t = fp8_division_transpose(
                    weight, self.qargs.group_size, self.fwobits["fwbit"]
                )
                setattr(self, f"{weight_name}_fp8_scale", weight_s)
                return weight_s
            else:
                # print(f"{weight_name} uses first microbatch")
                weight_fp8, weight_s, weight_fp8_t = fp8_division_transpose(
                    weight, self.qargs.group_size, self.fwobits["fwbit"]
                )
                setattr(self, f"{weight_name}_fp8", weight_fp8)
                setattr(self, f"{weight_name}_fp8_t", weight_fp8_t)
                setattr(self, f"{weight_name}_fp8_scale", weight_s)
                return weight_fp8, weight_fp8_t, weight_s
        else:
            if self.qargs.weight_memory_efficient:
                return getattr(self, f"{weight_name}_fp8_scale")
            else:
                return (
                    getattr(self, f"{weight_name}_fp8"),
                    getattr(self, f"{weight_name}_fp8_t"),
                    getattr(self, f"{weight_name}_fp8_scale"),
                )

    def forward(self, x):
        pass


# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import math
import struct

import numpy as np
import torch
import triton
import triton.language as tl
from triton.language.extra.cuda import libdevice


def floatExMy_quantize_triton(x, e_bit, m_bit, stochastic):
    n_elements = x.numel()
    grid = lambda meta: (triton.cdiv(n_elements, meta["BLOCK_SIZE"]),)
    y = torch.zeros_like(x)

    if x.dtype in [torch.bfloat16, torch.float32]:
        if stochastic:
            noise = x.new(x.shape).uniform_(-0.5, 0.5)
            _floatExMy_stochastic_quantize_kernel[grid](x, noise, y, n_elements, e_bit, m_bit)
        else:
            _floatExMy_quantize_kernel[grid](x, y, n_elements, e_bit, m_bit)
    else:
        raise NotImplementedError(f"Other data format {x.dtype} for float quantization triton")

    return y


@triton.autotune(
    configs=[
        # triton.Config({'BLOCK_SIZE': 4,}, num_warps=4),
        triton.Config(
            {
                "BLOCK_SIZE": 1024,
            },
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE": 2048,
            },
            num_stages=1,
        ),
    ],
    key=["n_elements"],
)
@triton.jit
def _floatExMy_quantize_kernel(
    x_ptr,
    output_ptr,
    n_elements,
    e_bit,
    m_bit,
    BLOCK_SIZE: tl.constexpr,
):
    if isinstance(e_bit, tl.constexpr):
        ebit = e_bit.value
    else:
        ebit = e_bit

    if isinstance(m_bit, tl.constexpr):
        mbit = m_bit.value
    else:
        mbit = m_bit

    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask)

    x = x.to(tl.float32)
    sign = 1 - 2 * libdevice.signbit(x)
    x_abs = tl.abs(x)
    Elow = -tl.exp2((ebit - 1).to(tl.float32)) + 2
    Ehigh = tl.exp2((ebit - 1).to(tl.float32))
    Mhigh = tl.exp2(mbit.to(tl.float32))
    expo = tl.floor(tl.log2(x_abs))
    expo = tl.clamp(expo, min=Elow, max=Ehigh)
    mant = x_abs / tl.exp2(expo)

    mant_int = tl.floor(mant)
    mant_frac = mant - mant_int
    mant_frac = mant_frac * Mhigh
    # mant_frac = mant_frac + noise
    mant_frac = libdevice.round(mant_frac)

    mant_q = mant_int + mant_frac / Mhigh
    y = sign * tl.exp2(expo) * mant_q
    y = y.to(x_ptr.dtype.element_ty)

    tl.store(output_ptr + offsets, y, mask=mask)


@triton.autotune(
    configs=[
        # triton.Config({'BLOCK_SIZE': 4,}, num_warps=4),
        triton.Config(
            {
                "BLOCK_SIZE": 1024,
            },
            num_warps=4,
        ),
        triton.Config(
            {
                "BLOCK_SIZE": 2048,
            },
            num_stages=1,
        ),
    ],
    key=["n_elements"],
)
@triton.jit
def _floatExMy_stochastic_quantize_kernel(
    x_ptr,
    noise_ptr,
    output_ptr,
    n_elements,
    e_bit,
    m_bit,
    BLOCK_SIZE: tl.constexpr,
):
    if isinstance(e_bit, tl.constexpr):
        ebit = e_bit.value
    else:
        ebit = e_bit

    if isinstance(m_bit, tl.constexpr):
        mbit = m_bit.value
    else:
        mbit = m_bit

    pid = tl.program_id(axis=0)
    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements
    x = tl.load(x_ptr + offsets, mask=mask)
    noise = tl.load(noise_ptr + offsets, mask=mask)

    x = x.to(tl.float32)
    sign = 1 - 2 * libdevice.signbit(x)
    x_abs = tl.abs(x)
    Elow = -tl.exp2((ebit - 1).to(tl.float32)) + 2
    Ehigh = tl.exp2((ebit - 1).to(tl.float32))
    Mhigh = tl.exp2(mbit.to(tl.float32))
    expo = tl.floor(tl.log2(x_abs))
    expo = tl.clamp(expo, min=Elow, max=Ehigh)
    mant = x_abs / tl.exp2(expo)

    mant_int = tl.floor(mant)
    mant_frac = mant - mant_int
    mant_frac = mant_frac * Mhigh
    mant_frac = mant_frac + noise
    mant_frac = libdevice.round(mant_frac)

    mant_q = mant_int + mant_frac / Mhigh
    y = sign * tl.exp2(expo) * mant_q
    y = y.to(x_ptr.dtype.element_ty)

    tl.store(output_ptr + offsets, y, mask=mask)



import torch
from transformers import AutoModelForCausalLM
from coat.activation.fake_quantization.quantize_function import extract_bit, block_cut, block_quant, block_reshape

# Fix the random seed
torch.manual_seed(0)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False


def eval_quant(input, precision, block_size=-1, symm=True, stochastic=False, epsilon=1e-12, apply_quantize=True, layer_name="test"):
    """
    Evaluate quantization performance for different precision formats and block sizes.
    
    Args:
        input: Input tensor to quantize
        precision: Quantization precision format (e.g., "E4M3", "MXE4M3", "NVE4M3")
        block_size: Block size for group quantization. Use -1 for per-tensor quantization.
        symm: Whether to use symmetric quantization
        stochastic: Whether to use stochastic rounding
        epsilon: Small value to avoid division by zero
        apply_quantize: Whether to apply quantization
        layer_name: Name of the layer for debugging
        
    Returns:
        tuple: (Quantized tensor, Quantized values, Scale factors, MSE)
    """
    # Cut the input into blocks
    Binput = block_cut(input, 1 if block_size > 0 else -1, block_size)
    
    # Apply quantization
    RQInput, q_values, scale_factors = block_quant(
        Binput, symm, precision, stochastic, epsilon, apply_quantize, layer_name
    )
    
    # Reshape back to original shape
    RQInput = block_reshape(RQInput, input, 1 if block_size > 0 else -1, block_size)
    
    # Calculate mean squared error
    mse = (input - RQInput).pow(2).mean().item()
    
    # Print results
    quant_type = "Per-tensor" if block_size <= 0 else f"Group={block_size}"
    print(f'MSE for {precision} {quant_type}: {mse}')
    
    return RQInput, q_values, scale_factors, mse

if __name__ == "__main__":
    input = torch.randn(4096, 4096).cuda()

    # Simulate outlier
    input[:, 100] *= 500
    input[100, :] *= 500

    # Test parameters
    stochastic = False
    epsilon = 1e-12
    apply_quantize = True
    layer_name = "test"
    symm = True
    
    print("\n=== Testing Per-tensor Quantization of E4M3 ===")
    precision, block_size = "E4M3", -1
    
    # Use the evaluate_quantization function
    RQInput_E4M3_tensor, q_e4m3, scale_e4m3, mse_E4M3_tensor = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )
    
    print("\n=== Testing Group = 128 Quantization of E4M3 ===")
    precision, block_size = "E4M3", 128
    RQInput_E4M3_group128, q_e4m3, scale_e4m3, mse_E4M3_group128 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )
    
    print("\n=== Testing Group = 32 Quantization of MXE5M1 ===")
    precision, block_size = "MXE5M1", 32
    RQInput_MXE5M1_group32, q_mx_e5m1, scale_mx_e5m1, mse_MXE5M1_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )
    
    print("\n=== Testing Group = 32 Quantization of MXE5M1_plus ===")
    precision, block_size = "MXE5M1_plus", 32
    RQInput_MXE5M1_plus_group32, q_mx_e5m1_plus, scale_mx_e5m1_plus, mse_MXE5M1_plus_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )
    
    print("\n=== Testing Group = 32 Quantization of MXE4M3 ===")
    precision, block_size = "MXE4M3", 32
    RQInput_MXE4M3_group32, q_mx_e4m3, scale_mx_e4m3, mse_MXE4M3_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )
    
    print("\n=== Testing Group = 32 Quantization of MXE3M2 ===")
    precision, block_size = "MXE3M2", 32
    RQInput_MXE3M2_group32, q_mx_e3m2, scale_mx_e3m2, mse_MXE3M2_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 32 Quantization of MXE2M3 ===")
    precision, block_size = "MXE2M3", 32
    RQInput_MXE2M3_group32, q_mx_e2m3, scale_mx_e2m3, mse_MXE2M3_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 32 Quantization of MXE2M1 ===")
    precision, block_size = "MXE2M1", 32
    RQInput_MXE2M1_group32, q_mx_e2m1, scale_mx_e2m1, mse_MXE2M1_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 32 Quantization of NVE5M1 ===")
    precision, block_size = "NVE5M1", 32
    RQInput_NVE5M1_group32, q_nv_e5m1, scale_nv_e5m1, mse_NVE5M1_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 32 Quantization of NVE5M1_plus ===")
    precision, block_size = "NVE5M1_plus", 32
    RQInput_NVE5M1_plus_group32, q_nv_e5m1_plus, scale_nv_e5m1_plus, mse_NVE5M1_plus_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 16 Quantization of NVE2M1 ===")
    precision, block_size = "NVE2M1", 16
    RQInput_NVE2M1_group16, q_nv_e2m1, scale_nv_e2m1, mse_NVE2M1_group16 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )

    print("\n=== Testing Group = 32 Quantization of NVE2M1_plus ===")
    precision, block_size = "NVE2M1_plus", 32
    RQInput_NVE2M1_plus_group32, q_nv_e2m1_plus, scale_nv_e2m1_plus, mse_NVE2M1_plus_group32 = eval_quant(
        input, precision, block_size, symm, stochastic, epsilon, apply_quantize, layer_name
    )



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

# Pass unit test

import unittest

import torch
from .test_utils import (check_similarity, dequantize_tensor, quantize_tensor,
                        random_tensor_generator)

from coat.activation.real_quantization.func_layernorm_noparam import (
    fp8_layernorm_noparam_backward, fp8_layernorm_noparam_forward)
from coat.activation.real_quantization.func_layernorm_param import (
    fp8_layernorm_param_backward, fp8_layernorm_param_forward)
from coat.activation.real_quantization.func_rmsnorm import (
    fp8_rmsnorm_backward, fp8_rmsnorm_forward)


def _test_layer_norm_noparam(x, qx, sx, g, BS, SL, CDIM, QB, eps=1e-5):
    """
    Test the forward and backward of LayerNorm (no param)
    Forward: input is per-group quant, output is per-tensor quanti
    Backward: input and output are full precision
    """
    x = x.clone().requires_grad_(True)

    # Forward
    w_shape = (CDIM,)
    torch_layer_norm = torch.nn.functional.layer_norm
    _output_torch = torch_layer_norm(x, w_shape, eps=eps)
    output_torch_fwd, _, _ = quantize_tensor(_output_torch, BS, SL, CDIM, QB, qx.dtype, quant_type="per_block")

    x_triton, s_triton, x_t_triton, (mean, rstd, num_warps) = fp8_layernorm_noparam_forward(qx, sx, QB, eps)
    output_triton_fwd = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)

    # Backward
    _output_torch.backward(g)
    output_torch_bwd = x.grad.to(torch.bfloat16)

    output_triton_bwd = fp8_layernorm_noparam_backward(qx, sx, g, QB, mean, rstd, num_warps)

    return output_torch_fwd, output_triton_fwd, output_torch_bwd, output_triton_bwd


def _test_layer_norm_param(x, qx, sx, g, w, b, BS, SL, CDIM, QB, eps=1e-5):
    """
    Test the forward and backward of LayerNorm (no param)
    Forward: input is per-group quant, output is per-tensor quanti
    Backward: input and output are full precision
    """
    x = x.clone().requires_grad_(True)

    # Forward
    w_shape = (CDIM,)
    torch_layer_norm = torch.nn.functional.layer_norm
    _output_torch = torch_layer_norm(x, w_shape, weight=w, bias=b, eps=eps)
    output_torch_fwd, _, _ = quantize_tensor(_output_torch, BS, SL, CDIM, QB, qx.dtype, quant_type="per_block")

    x_triton, s_triton, x_t_triton, (_, _, mean, rstd, num_warps) = fp8_layernorm_param_forward(qx, sx, w, b, QB, eps)
    output_triton_fwd = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)

    # Backward
    _output_torch.backward(g)
    output_torch_bwd = x.grad.to(torch.bfloat16)
    output_w_torch_bwd = w.grad
    output_b_torch_bwd = b.grad

    output_triton_bwd, output_w_triton_bwd, output_b_triton_bwd = fp8_layernorm_param_backward(qx, sx, g, QB, w, b, mean, rstd, num_warps)

    return (
        output_torch_fwd, output_triton_fwd, 
        output_torch_bwd, output_triton_bwd,
        output_w_torch_bwd, output_w_triton_bwd,
        output_b_torch_bwd, output_b_triton_bwd,
    )


def _test_rms_norm(x, qx, sx, g, w, BS, SL, CDIM, QB, eps=1e-5):
    """
    Test the forward and backward of RMSNorm (no param)
    Forward: input is per-group quant, output is per-tensor quanti
    Backward: input and output are full precision
    """
    x = x.clone().requires_grad_(True)

    # Forward
    w_shape = (CDIM,)
    torch_rms_norm = torch.nn.functional.rms_norm
    _output_torch = torch_rms_norm(x, w_shape, w, eps=eps)
    output_torch_fwd, _, _ = quantize_tensor(_output_torch, BS, SL, CDIM, QB, qx.dtype, quant_type="per_block")

    x_triton, s_triton, x_t_triton, (_, rstd, num_warps) = fp8_rmsnorm_forward(qx, sx, w, QB, eps)
    output_triton_fwd = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)

    # Backward
    _output_torch.backward(g)
    output_torch_bwd = x.grad.to(torch.bfloat16)
    output_w_torch_bwd = w.grad

    output_triton_bwd, output_w_triton_bwd = fp8_rmsnorm_backward(qx, sx, g, w, rstd, QB, num_warps)

    return (
        output_torch_fwd,
        output_triton_fwd,
        output_torch_bwd,
        output_triton_bwd,
        output_w_torch_bwd,
        output_w_triton_bwd,
    )


class TestNorm(unittest.TestCase):
    """
    x: reconstructed activation (quantize it then dequantize)
    qx: quantized activation
    sx: scale factor of activation

    g: reconstrcted gradient (quantize it then dequantize)
    qg: quantized gradient
    sg: scale factor of gradient
    """

    # Test the forward of LayerNorm
    def test_layernorm_noparam(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        output_torch_fwd, output_triton_fwd, output_torch_bwd, output_triton_bwd = _test_layer_norm_noparam(
            x, qx, sx, g, BS, SL, CDIM, QB
        )

    # Test the forward of LayerNorm
    def test_layernorm_param(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        w, b = [torch.rand((CDIM,), device="cuda", requires_grad=True) for _ in range(2)]
        (
            output_torch_fwd, output_triton_fwd, 
            output_torch_bwd, output_triton_bwd,
            output_w_torch_bwd, output_w_triton_bwd,
            output_b_torch_bwd, output_b_triton_bwd,
        ) = _test_layer_norm_param(x, qx, sx, g, w, b, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch_fwd, output_triton_fwd))
        self.assertTrue(check_similarity(output_torch_bwd, output_triton_bwd))
        self.assertTrue(torch.allclose(output_w_torch_bwd + 1e-8, output_w_triton_bwd + 1e-8, 1e-3))
        self.assertTrue(torch.allclose(output_b_torch_bwd + 1e-8, output_b_triton_bwd + 1e-8, 1e-3))

    # Test the forward of RMSNorm
    def test_rmsnorm(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        w = torch.rand((CDIM,), device="cuda", requires_grad=True)
        (
            output_torch_fwd,
            output_triton_fwd,
            output_torch_bwd,
            output_triton_bwd,
            output_w_torch_bwd,
            output_w_triton_bwd,
        ) = _test_rms_norm(x, qx, sx, g, w, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        self.assertTrue(check_similarity(output_torch_fwd, output_triton_fwd))
        self.assertTrue(check_similarity(output_torch_bwd, output_triton_bwd))
        self.assertTrue(torch.allclose(output_w_torch_bwd + 1e-8, output_w_triton_bwd + 1e-8, 1e-3))


if __name__ == "__main__":
    torch.manual_seed(0)
    torch.set_printoptions(sci_mode=False, linewidth=200)
    unittest.main()



import torch
from ..activation.deepseek.fp8linear import FP8DeepSeekLinear

def fp8_deepseek_linear_replacer(model, linear_names: list = []):
    if linear_names == []:
        linear_names.extend(["q_proj", "k_proj", "v_proj", "o_proj", "gate_proj", "up_proj", "down_proj"])
        linear_names.extend(["attn_out", "ff_out", "att_proj", "ff_proj"])

    for name, module in model.named_modules():
        if isinstance(module, torch.nn.Linear) and any(key in name for key in linear_names):
            new_module = FP8DeepSeekLinear(module.in_features, module.out_features)
            new_module.weight.data.copy_(module.weight.data)
            if module.bias is not None:
                new_module.bias.data.copy_(module.bias.data)
                
            setattr(model, name, new_module)
    return model


# In this file I only add the logic in line 453 to 456. The rest remains unchanged compared with the original Trainer Class.

import os
import math
import time
import torch
import shutil
from packaging import version
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import torch.nn as nn

from transformers import Trainer
from transformers.debug_utils import DebugOption, DebugUnderflowOverflow
from torch.utils.data import DataLoader, Dataset, IterableDataset, RandomSampler, SequentialSampler
from transformers.trainer_pt_utils import get_model_param_count
from transformers.integrations import hp_params
from transformers.trainer_utils import (
    has_length,
    speed_metrics,
    HPSearchBackend,
    TrainOutput
)
from transformers.training_args import OptimizerNames, ParallelMode, TrainingArguments
from transformers.trainer_callback import (
    TrainerState,
    ExportableState,
)
from transformers.integrations.tpu import tpu_spmd_dataloader
from transformers.integrations.deepspeed import deepspeed_init, deepspeed_load_checkpoint, is_deepspeed_available
from transformers.utils import (
    is_accelerate_available,
    is_apex_available,
    is_bitsandbytes_available,
    is_datasets_available,
    is_galore_torch_available,
    is_grokadamw_available,
    is_in_notebook,
    is_ipex_available,
    is_liger_kernel_available,
    is_lomo_available,
    is_peft_available,
    is_safetensors_available,
    is_sagemaker_dp_enabled,
    is_sagemaker_mp_enabled,
    is_schedulefree_available,
    is_torch_compile_available,
    is_torch_mlu_available,
    is_torch_mps_available,
    is_torch_musa_available,
    is_torch_neuroncore_available,
    is_torch_npu_available,
    is_torch_xla_available,
    is_torch_xpu_available,
    is_torchao_available,
    logging
)

if is_apex_available():
    from apex import amp

if is_datasets_available():
    import datasets

if is_torch_xla_available():
    import torch_xla.core.xla_model as xm
    import torch_xla.debug.metrics as met
    from torch_xla import __version__ as XLA_VERSION

    IS_XLA_FSDPV2_POST_2_2 = version.parse(XLA_VERSION) >= version.parse(XLA_FSDPV2_MIN_VERSION)
    if IS_XLA_FSDPV2_POST_2_2:
        import torch_xla.distributed.spmd as xs
        import torch_xla.runtime as xr
else:
    IS_XLA_FSDPV2_POST_2_2 = False


if is_sagemaker_mp_enabled():
    import smdistributed.modelparallel.torch as smp
    from smdistributed.modelparallel import __version__ as SMP_VERSION

    IS_SAGEMAKER_MP_POST_1_10 = version.parse(SMP_VERSION) >= version.parse("1.10")

    from .trainer_pt_utils import smp_forward_backward, smp_forward_only, smp_gather, smp_nested_concat
else:
    IS_SAGEMAKER_MP_POST_1_10 = False


if is_safetensors_available():
    import safetensors.torch

if is_peft_available():
    from peft import PeftModel


if is_accelerate_available():
    from accelerate import Accelerator, skip_first_batches
    from accelerate import __version__ as accelerate_version
    from accelerate.utils import (
        DistributedDataParallelKwargs,
        DistributedType,
        GradientAccumulationPlugin,
        load_fsdp_model,
        load_fsdp_optimizer,
        save_fsdp_model,
        save_fsdp_optimizer,
    )

    DATA_SAMPLERS = [RandomSampler]
    if version.parse(accelerate_version) > version.parse("0.23.0"):
        from accelerate.data_loader import SeedableRandomSampler

        DATA_SAMPLERS += [SeedableRandomSampler]

    if is_deepspeed_available():
        from accelerate.utils import DeepSpeedSchedulerWrapper

if is_accelerate_available("0.28.0"):
    from accelerate.utils import DataLoaderConfiguration


from .utils._fp8manager import FP8Manager
from .optimizer.fp8_adamw import CoatAdamW

logger = logging.get_logger(__name__)

def get_local_rank() -> int:
    return int(os.environ.get("LOCAL_RANK") or 0)

class CoatFP8Trainer(Trainer):
    def __init__(self, coat_args=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qargs = coat_args  # Additional parameter

    def create_optimizer(self):
        """
        Setup the FP8 optimizer.
        """
        if get_local_rank() == 0:
            print("Setup the COAT FP8 optimizer")
        opt_model = self.model

        if self.optimizer is None:
            decay_parameters = self.get_decay_parameter_names(opt_model)
            optimizer_grouped_parameters = [
                {
                    "params": [
                        p for n, p in opt_model.named_parameters() if (n in decay_parameters and p.requires_grad)
                    ],
                    "weight_decay": self.args.weight_decay,
                },
                {
                    "params": [
                        p for n, p in opt_model.named_parameters() if (n not in decay_parameters and p.requires_grad)
                    ],
                    "weight_decay": 0.0,
                },
            ]

            optimizer_kwargs2 = {
                "lr": self.args.learning_rate,
                "betas": (self.args.adam_beta1, self.args.adam_beta2),
                "eps": self.args.adam_epsilon,
            }
            if self.qargs is not None:
                optimizer_kwargs2["qargs"] = self.qargs
                
            self.optimizer = CoatAdamW(optimizer_grouped_parameters, **optimizer_kwargs2)

        return self.optimizer
    
    # Possibly, put this into trainers's logic
    # if total_batched_samples % args.gradient_accumulation_steps == 0:
    #     FP8Manager.is_first_microbatch = True
    # else:
    #     FP8Manager.is_first_microbatch = False



# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

import os
import time
from copy import deepcopy
from dataclasses import dataclass

import matplotlib.pyplot as plt
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd.function import Function

from ..utils import quant_get_local_rank
from ..real_quantization._quantize import fp8_quantize
from ..real_quantization._quantize_transpose import fp8_quantize_transpose
from ..real_quantization._quantize_perblock import fp8_quantize_perblock
from ..real_quantization._quantize_perblock_transpose import fp8_quantize_perblock_transpose
from ..real_quantization._quantize_perblock_transpose_pergroup import fp8_quantize_perblock_transpose_pergroup
from .linear import fp8_deepseek_linear_backward, fp8_deepseek_linear_forward

import deep_gemm

@dataclass
class DefaultArgs:
    fabit: str = "E4M3"
    fwbit: str = "E4M3"
    bobit: str = "E4M3"


class FP8DeepSeekLinear(nn.Linear):
    def __init__(self, in_features, out_features, bias=True, device=None, args=None, layer_idx=0):
        super().__init__(in_features, out_features, bias, device)

        if args is None:  # I do not want to pass a new argument to OLMo so just use this method
            args = DefaultArgs()
        self.args = deepcopy(args)

        if quant_get_local_rank() == 0:
            print(f"[qlinear debug] Apply QLinear, {layer_idx}")

        self.layer_idx = layer_idx
        self.layer_name = None

    def forward(self, Input):
        if self.training:
            # if False:
            output = _FP8DeepSeekLinear.apply(Input, self.weight, self.bias, self.args, self.layer_name)
        else:
            output = F.linear(Input, self.weight, self.bias)

        return output

class _FP8DeepSeekLinear(Function):
    @staticmethod
    @torch.amp.custom_fwd(device_type="cuda")
    def forward(ctx, input, weight, bias, args, layer_name):
        (Qinput, Iscale), (Qinput_t, ITscale) = fp8_quantize_transpose(input, 128, args.fabit, scale_dtype=torch.float32)
        # To match deepseek's kernel requirement
        Iscale = Iscale.t().contiguous().t()
        ITscale = ITscale.t().contiguous().t()

        (Qweight, Wscale) = fp8_quantize_perblock(weight, 128, args.fwbit, scale_dtype=torch.float32)

        ctx.save_for_backward(Qinput_t, ITscale, weight, bias)
        ctx.utils = args, layer_name, input.shape, weight.shape
        
        fc_output = fp8_deepseek_linear_forward(Qinput, Iscale, Qweight, Wscale, 0, bias)

        output_shape = list(input.shape)
        output_shape[-1] = weight.shape[0]
        
        fc_output = fc_output.view(output_shape)

        return fc_output

    @staticmethod
    @torch.amp.custom_bwd(device_type="cuda")
    def backward(ctx, grad_output):
        Qinput_t, ITscale, weight, bias = ctx.saved_tensors
        args, layer_name, input_shape, weight_shape = ctx.utils
        grad_output_shape = grad_output.shape
        
        # Flatten the gradient
        grad_output = grad_output.view(-1, grad_output.shape[-1])

        # For DGrad
        (Qgrad_output_pb, Gscale_pb), (Qgrad_output_t, Gscale_t), (Qgrad_output_pg, Gscale_pg) = \
            fp8_quantize_perblock_transpose_pergroup(grad_output, 128, args.bobit, scale_dtype=torch.float32, only_transposed=True)

        Gscale_pg = Gscale_pg.t().contiguous().t()

        Qweight_t, Wscale_t = fp8_quantize_perblock_transpose(weight, 128, args.fwbit, scale_dtype=torch.float32, only_transposed=True)

        grad_input, grad_weight = fp8_deepseek_linear_backward(
            Qinput_t,
            ITscale,
            Qgrad_output_pg,
            Gscale_pg,
            Qgrad_output_t,
            Gscale_t,
            Qweight_t,
            Wscale_t,
        )

        if bias is not None:
            grad_bias = grad_output.reshape(-1, grad_output.shape[-1]).sum(0)
        else:
            grad_bias = None

        grad_input = grad_input.view(input_shape)

        return grad_input, grad_weight, grad_bias, None, None


# Copyright 2024 NVIDIA CORPORATION & AFFILIATES
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0

# PASS the unit test

import unittest

import torch
from test_utils import (check_similarity, dequantize_tensor, quantize_tensor,
                        random_tensor_generator)

from coat.activation.real_quantization.common import FP8_MAX_VALUE
from coat.activation.real_quantization.gelu_bwd import fp8_gelu_backward
from coat.activation.real_quantization.gelu_bwd_legacy import \
    fp8_gelu_backward_legacy
from coat.activation.real_quantization.gelu_fwd import fp8_gelu_forward
from coat.activation.real_quantization.mul_bwd import fp8_mul_backward
from coat.activation.real_quantization.mul_bwd_legacy import \
    fp8_mul_backward_legacy
from coat.activation.real_quantization.mul_fwd import fp8_mul_forward
from coat.activation.real_quantization.silu_bwd import fp8_silu_backward
from coat.activation.real_quantization.silu_bwd_legacy import \
    fp8_silu_backward_legacy
from coat.activation.real_quantization.silu_fwd import fp8_silu_forward


def _test_gelu_fwd(x, qx, sx, BS, SL, CDIM, QB):
    """
    Tests the forward pass of the GELU activation function,
    comparing the results between standard PyTorch and a custom FP8-based implementation.

    Returns:
    - output_torch (torch.Tensor): GELU output computed using PyTorch.
                                   We also quantize the output since we use fp8 precision flow.
    - output_triton (torch.Tensor): GELU output computed using FP8-based implementation.
    """
    torch_gelu = torch.nn.GELU()
    output_torch = torch_gelu(x)
    output_torch, _, _ = quantize_tensor(output_torch, BS, SL, CDIM, QB, qx.dtype, quant_type="per_block")

    x_triton, s_triton, x_t_triton = fp8_gelu_forward(qx, sx, QB)
    output_triton = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)

    return output_torch, output_triton


def _test_gelu_bwd(x, qx, sx, g, BS, SL, CDIM, QB, fp8type):
    """
    Tests the backward pass of the GELU activation function,
    The output is per-tensor quantized, since it should follows a linear layer.
    """
    x = x.clone().requires_grad_(True)

    torch_gelu = torch.nn.GELU()
    output_torch = torch_gelu(x)
    output_torch.backward(g)
    grad_torch, _, __ = quantize_tensor(
        x.grad, BS, SL, CDIM, QB, fp8type, quant_type="per_block"
    )  # .to(bfloat16) since gradients are of bfloat16

    g_triton, s_triton = fp8_gelu_backward(qx, sx, g, QB, fp8type)
    grad_triton = dequantize_tensor(
        g_triton.unsqueeze(0), s_triton.unsqueeze(0), 1, SL, CDIM, QB
    )  # dequantize per-tensor quantization

    return grad_torch, grad_triton


def _test_gelu_bwd_legacy(x, qx, sx, g, qg, sg, BS, SL, CDIM, QB):
    """
    Tests the backward pass of the GELU activation function,
    The output is per-tensor quantized, since it should follows a linear layer.
    """
    x = x.clone().requires_grad_(True)

    torch_gelu = torch.nn.GELU()
    output_torch = torch_gelu(x)
    output_torch.backward(g)
    grad_torch, _, __ = quantize_tensor(
        x.grad, BS, SL, CDIM, QB, qg.dtype, quant_type="per_block"
    )  # .to(bfloat16) since gradients are of bfloat16

    g_triton, s_triton = fp8_gelu_backward_legacy(qx, sx, qg, sg, QB)
    grad_triton = dequantize_tensor(g_triton, s_triton, BS, SL, CDIM, QB)  # dequantize per-tensor quantization

    return grad_torch, grad_triton


def _test_silu_fwd(x, qx, sx, BS, SL, CDIM, QB):
    """
    Tests the forward pass of the SiLU activation function,
    """
    torch_silu = torch.nn.SiLU()
    output_torch = torch_silu(x)
    output_torch, _, _ = quantize_tensor(output_torch, BS, SL, CDIM, QB, qx.dtype)

    x_triton, s_triton = fp8_silu_forward(qx, sx, QB)
    output_triton = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)
    return output_torch, output_triton


def _test_silu_bwd(x, qx, sx, g, BS, SL, CDIM, QB, fp8type):
    """
    Tests the backward pass of the SiLU activation function,
    The output is per-tensor full precision, since it should follows a linear layer. We do not quantize it right not, and require to quantize it afterwards.
    """
    x = x.clone().requires_grad_(True)

    torch_silu = torch.nn.SiLU()
    output_torch = torch_silu(x)
    output_torch.backward(g)
    grad_torch = x.grad.to(torch.bfloat16)
    grad_torch, _, __ = quantize_tensor(
        x.grad, BS, SL, CDIM, QB, fp8type, quant_type="per_block"
    )  # .to(bfloat16) since gradients are of bfloat16

    grad_triton, _ = fp8_silu_backward(qx, sx, g, QB, fp8type)

    return grad_torch, grad_triton


def _test_silu_bwd_legacy(x, qx, sx, g, qg, sg, BS, SL, CDIM, QB):
    """
    Tests the backward pass of the SiLU activation function,
    The output is per-tensor full precision, since it should follows a linear layer. We do not quantize it right not, and require to quantize it afterwards.
    """
    x = x.clone().requires_grad_(True)

    torch_silu = torch.nn.SiLU()
    output_torch = torch_silu(x)
    output_torch.backward(g)
    grad_torch = x.grad.to(torch.bfloat16)

    grad_triton, _ = fp8_silu_backward_legacy(qx, sx, qg, sg, QB)

    return grad_torch, grad_triton


def _test_mul_fwd(x1, qx1, sx1, x2, qx2, sx2, BS, SL, CDIM, QB):
    """
    Tests the forward pass of the Element-wise Multiplication function,
    X1 represents Gate Proj, and X2 represents Up Proj.
    The output is per-tensor quantized, since it should follows a linear layer.
    """
    output_torch = x1 * x2
    output_torch, _, _ = quantize_tensor(output_torch, BS, SL, CDIM, QB, qx1.dtype, quant_type="per_block")

    x_triton, s_triton, x_t_triton = fp8_mul_forward(qx1, sx1, qx2, sx2, QB)
    output_triton = dequantize_tensor(x_triton, s_triton, BS, SL, CDIM, QB)

    return output_torch, output_triton


def _test_mul_bwd(x1, qx1, sx1, x2, qx2, sx2, g, BS, SL, CDIM, QB, fp8type):
    """
    Tests the backward pass of the Element-wise Multiplication function,
    We return the gradient of Gate Proj and Up Proj. G1 is per-group quantized, and G2 is per-tensor quantized.
    """
    x1 = x1.clone().requires_grad_(True)
    x2 = x2.clone().requires_grad_(True)
    output_torch = x1 * x2
    output_torch.backward(g)
    output_torch1 = x1.grad.to(torch.bfloat16)
    output_torch2, _, __ = quantize_tensor(x2.grad, BS, SL, CDIM, QB, fp8type, quant_type="per_block")

    g1_triton, (output_triton2, _) = fp8_mul_backward(qx1, sx1, qx2, sx2, g, QB, fp8type)
    output_triton1 = g1_triton

    return output_torch1, output_triton1, output_torch2, output_triton2


def _test_mul_bwd_legacy(x1, qx1, sx1, x2, qx2, sx2, g, qg, sg, BS, SL, CDIM, QB):
    """
    Tests the backward pass of the Element-wise Multiplication function,
    We return the gradient of Gate Proj and Up Proj. G1 is per-group quantized, but G2 is per-tensor full precision.
    """
    x1 = x1.clone().requires_grad_(True)
    x2 = x2.clone().requires_grad_(True)
    output_torch = x1 * x2
    output_torch.backward(g)
    output_torch1, _, __ = quantize_tensor(
        x1.grad, BS, SL, CDIM, QB, qg.dtype
    )  # NOTE: qg.dtype is e5m2, while qx uses e4m3
    output_torch2 = x2.grad.to(torch.bfloat16)

    g1_triton, s1_triton, g2_triton, s2_triton = fp8_mul_backward_legacy(qx1, sx1, qx2, sx2, qg, sg, QB)
    output_triton1 = dequantize_tensor(g1_triton, s1_triton, BS, SL, CDIM, QB)
    output_triton2 = g2_triton

    return output_torch1, output_triton1, output_torch2, output_triton2


class TestActivation(unittest.TestCase):
    """
    x: reconstructed activation (quantize it then dequantize)
    qx: quantized activation
    sx: scale factor of activation

    g: reconstrcted gradient (quantize it then dequantize)
    qg: quantized gradient
    sg: scale factor of gradient
    """

    # Test the forward of GELU
    def test_gelu_fwd(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        output_torch, output_triton = _test_gelu_fwd(x, qx, sx, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the backward of GELU
    def test_gelu_bwd(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)

        output_torch, output_triton = _test_gelu_bwd(x, qx, sx, g, BS, SL, CDIM, QB, fp8type_grad)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the backward of GELU
    def test_gelu_bwd_legacy(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, qg, sg = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)
        output_torch, output_triton = _test_gelu_bwd_legacy(x, qx, sx, g, qg, sg, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the forward of SiLU
    def test_silu_fwd(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        output_torch, output_triton = _test_silu_fwd(x, qx, sx, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the backward of SiLU
    def test_silu_bwd(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)
        output_torch, output_triton = _test_silu_bwd(x, qx, sx, g, BS, SL, CDIM, QB, fp8type_grad)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the backward of SiLU
    def test_silu_bwd_legacy(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x, qx, sx = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, qg, sg = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)
        output_torch, output_triton = _test_silu_bwd_legacy(x, qx, sx, g, qg, sg, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the forward of element-wise multiplication
    def test_mul_fwd(self):
        BS, SL, CDIM, QB, fp8type = [4, 256, 2048, 16, torch.float8_e4m3fn]
        x1, qx1, sx1 = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        x2, qx2, sx2 = random_tensor_generator(BS, SL, CDIM, QB, fp8type)
        output_torch, output_triton = _test_mul_fwd(x1, qx1, sx1, x2, qx2, sx2, BS, SL, CDIM, QB)

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch, output_triton))

    # Test the backward of element-wise multiplication
    def test_mul_bwd(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x1, qx1, sx1 = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        x2, qx2, sx2 = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, _, _ = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)
        output_torch1, output_triton1, output_torch2, output_triton2 = _test_mul_bwd(
            x1, qx1, sx1, x2, qx2, sx2, g, BS, SL, CDIM, QB, fp8type_grad
        )

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch1, output_triton1))
        self.assertTrue(check_similarity(output_torch2, output_triton2))

    # Test the backward of element-wise multiplication
    def test_mul_bwd_legacy(self):
        BS, SL, CDIM, QB, fp8type_act, fp8type_grad = [4, 256, 2048, 16, torch.float8_e4m3fn, torch.float8_e5m2]
        x1, qx1, sx1 = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        x2, qx2, sx2 = random_tensor_generator(BS, SL, CDIM, QB, fp8type_act)
        g, qg, sg = random_tensor_generator(BS, SL, CDIM, QB, fp8type_grad)
        output_torch1, output_triton1, output_torch2, output_triton2 = _test_mul_bwd_legacy(
            x1, qx1, sx1, x2, qx2, sx2, g, qg, sg, BS, SL, CDIM, QB
        )

        # This is a very loose assert. Most values should be the same.
        # self.assertTrue(torch.allclose(output_torch + 1e-8, output_triton + 1e-8, 1e-2, 0.2))
        self.assertTrue(check_similarity(output_torch1, output_triton1))
        self.assertTrue(check_similarity(output_torch2, output_triton2))


if __name__ == "__main__":
    torch.manual_seed(0)
    torch.set_printoptions(sci_mode=False, linewidth=200)
    unittest.main()
